/**
 *
 * Copyright (c) 2005 Conitec Datasystems GmbH
 * Written 2005 by Peer Georgi <georgi@conitec.com>
 *
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for complete details.
 **
*/


#include "sc_UDP.h"
#include <at/include/at91rm9200-def.h>
#include <sc/sc_config.h>
#include <sc/sc.h>

// #define UDP_DEBUG_MODE   // be very verbose //

#define STD_GET_STATUS_ZERO           0x0080
#define STD_GET_STATUS_INTERFACE      0x0081
#define STD_GET_STATUS_ENDPOINT       0x0082

#define STD_CLEAR_FEATURE_ZERO        0x0100
#define STD_CLEAR_FEATURE_INTERFACE   0x0101
#define STD_CLEAR_FEATURE_ENDPOINT    0x0102

#define STD_SET_FEATURE_ZERO          0x0300
#define STD_SET_FEATURE_INTERFACE     0x0301
#define STD_SET_FEATURE_ENDPOINT      0x0302

#define STD_SET_ADDRESS               0x0500
#define STD_GET_DESCRIPTOR            0x0680
#define STD_SET_DESCRIPTOR            0x0700
#define STD_GET_CONFIGURATION         0x0880
#define STD_SET_CONFIGURATION         0x0900
#define STD_GET_INTERFACE             0x0A81
#define STD_SET_INTERFACE             0x0B01

#define LDBA_SET_CMD                  0xf121 // host -> device //

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

#define UDP_DEBUG_WARNING(Msg) { scDbg("UDP-Warning: "Msg); }
#define UDP_DEBUG_WARNING_HEX(i) { scDbgPrintHex32(i); }

#ifdef UDP_DEBUG_MODE
  #define UDP_DEBUG_V3(Msg) { scDbg(Msg); }
  #define UDP_DEBUG_V3_HEX(i) { scDbgPrintHex32(i); }
#else
  #define UDP_DEBUG_V3(Msg) 
  #define UDP_DEBUG_V3_HEX(i) 
#endif

const char devDescriptor[] = {
    /* Device descriptor */
    0x12,   // bLength
    0x01,   // bDescriptorType
    0x10,   // bcdUSBL
    0x01,   //
    0x02,   // bDeviceClass:    CDC class code
    0x00,   // bDeviceSubclass: CDC class sub code
    0x00,   // bDeviceProtocol: CDC Device protocol
    0x08,   // bMaxPacketSize0
    0xEB,   // idVendorL
    0x03,   //
    0x22,   // idProductL
    0x61,   //
    0x10,   // bcdDeviceL
    0x01,   //
    0x00,   // iManufacturer    // 0x01
    0x00,   // iProduct
    0x00,   // SerialNumber
    0x01    // bNumConfigs
};

const char cfgDescriptor[] = {
    /* ============== CONFIGURATION 1 =========== */
    /* Configuration 1 descriptor */
    0x09,   // CbLength
    0x02,   // CbDescriptorType
    0x20,   // CwTotalLength 2 EP + Control
    0x00,
    0x01,   // CbNumInterfaces
    0x01,   // CbConfigurationValue
    0x00,   // CiConfiguration
    0xC0,   // CbmAttributes 0xA0
    0x00,   // CMaxPower

    /* Data Class Interface Descriptor Requirement */
    0x09, // bLength
    0x04, // bDescriptorType
    0x00, // bInterfaceNumber
    0x00, // bAlternateSetting
    0x02, // bNumEndpoints
    0x00, // bInterfaceClass
    0x00, // bInterfaceSubclass
    0x00, // bInterfaceProtocol
    0x00, // iInterface

    /* First alternate setting */
    /* Endpoint 1 descriptor */
    0x07,   // bLength
    0x05,   // bDescriptorType
    0x01,   // bEndpointAddress, Endpoint 01 - OUT
    0x02,   // bmAttributes      BULK
    0x40,   // wMaxPacketSize
    0x00,
    0x00,   // bInterval

    /* Endpoint 2 descriptor */
    0x07,   // bLength
    0x05,   // bDescriptorType
    0x82,   // bEndpointAddress, Endpoint 02 - IN
    0x02,   // bmAttributes      BULK
    0x40,   // wMaxPacketSize
    0x00,
    0x00    // bInterval
};

unsigned short giCurrentConfig = 0;

/** PRIVATES **/

#define PRINT_REG(s, R) {scDbg(s": 0x");scDbgPrintHex32(R);scDbg("\n"); }
void __sc_udp_debug_printRegs()
  {
    scDbg("\n");
    PRINT_REG("UDP_FRM", AT91C_BASE_UDP->UDP_NUM);
    PRINT_REG("UDP_CSR0", AT91C_BASE_UDP->UDP_CSR[0]);
    PRINT_REG("UDP_FDR0", AT91C_BASE_UDP->UDP_FDR[0]);
    PRINT_REG("UDP_GSTAT", AT91C_BASE_UDP->UDP_GLBSTATE);
    PRINT_REG("UDP_FADDR", AT91C_BASE_UDP->UDP_FADDR);
    PRINT_REG("UDP_ISR", AT91C_BASE_UDP->UDP_ISR);
  }

void __sc_udp_handleTimeOut(AT91PS_UDP pUdp)
  {
//    __sc_udp_sendStall(pUdp);
//    pUdp->UDP_CSR[0] = 0;
  }

/**
  Waits for pUdp->UDP_CSR[0] & AT91C_UDP_TXPKTRDY == 0 and returns 1.
  Return Value
  If a time out has been occured the return value will be 0.
*/
char __sc_udp_waitTxReady(AT91PS_UDP pUdp)
  {
    unsigned int iTimeOut = 1024*1024;
    while ((pUdp->UDP_CSR[0] & AT91C_UDP_TXPKTRDY))
      {
        iTimeOut--;
        if (!(iTimeOut)) return (0); // timeout //
      }
    return (1);
  }

/**
  Waits for pUdp->UDP_CSR[0] & AT91C_UDP_TXCOMP == 1, clears the flag and returns 1.
  Return Value
  If a time out has been occured the return value will be 0.
*/
char __sc_udp_waitTxCompleate(AT91PS_UDP pUdp)
  {
    unsigned int iTimeOut = 1024*1024;
    while (1)
      {
        unsigned int iCSR = pUdp->UDP_CSR[0];
        if (iCSR & AT91C_UDP_TXCOMP)
          break;
        if (iCSR & AT91C_UDP_FORCESTALL)
          scDbg("Stall!\n");
        iTimeOut--;      
        if (!(iTimeOut)) return (0); // timeout //
      }
//     while (!(pUdp->UDP_CSR[0] & AT91C_UDP_TXCOMP))
//       {
// /*              if (pUdp->UDP_CSR[0] & AT91C_UDP_RX_DATA_BK0)
//               {
//             pUdp->UDP_CSR[0] &= ~(AT91C_UDP_RX_DATA_BK0);
//                   UDP_DEBUG_WARNING("transmition stopped by a status OUT (incoming data)\n");
//             return (0);
//           }*/
//         iTimeOut--;
//         if (!(iTimeOut)) return (0); // timeout //
//       }
    pUdp->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);
    return (1);
  }


#define     AT91_UDP_TXCOMP     (1 <<  0)       /* Generates IN packet with data previously written in DPR */
#define     AT91_UDP_RX_DATA_BK0    (1 <<  1)       /* Receive Data Bank 0 */
#define     AT91_UDP_RXSETUP    (1 <<  2)       /* Send STALL to the host */
#define     AT91_UDP_STALLSENT  (1 <<  3)       /* Stall Sent / Isochronous error (Isochronous endpoints) */
#define     AT91_UDP_TXPKTRDY   (1 <<  4)       /* Transmit Packet Ready */
#define     AT91_UDP_FORCESTALL (1 <<  5)       /* Force Stall */
#define     AT91_UDP_RX_DATA_BK1    (1 <<  6)       /* Receive Data Bank 1 */
#define     AT91_UDP_DIR        (1 <<  7)       /* Transfer Direction */
#define     AT91_UDP_EPTYPE     (7 <<  8)       /* Endpoint Type */

#define RX_DATA_READY   (AT91_UDP_RX_DATA_BK0 | AT91_UDP_RX_DATA_BK1)

#define SET_FX  (AT91C_UDP_TXPKTRDY)
#define CLR_FX  (RX_DATA_READY | AT91_UDP_RXSETUP | AT91_UDP_STALLSENT | AT91_UDP_TXCOMP)

/**
  Writes a frame to usb.
*/
void __sc_udp_write(AT91PS_UDP pUdp, const char *pData, unsigned int iLength)
  {
   unsigned int iPacketSize = 0;
   
   // accept previous (failed) transmitions (paranoya) //
   pUdp->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);

   UDP_DEBUG_V3("write: ");
   if (!(__sc_udp_waitTxReady(pUdp)))
     {
       UDP_DEBUG_WARNING("UDP-Write timeout stage 0\n");
       __sc_udp_handleTimeOut(pUdp);
       return;
     }
    do
    {
      unsigned int iCSR = pUdp->UDP_CSR[0];
      iPacketSize = MIN(iLength, 8);
      iLength -= iPacketSize;
      while (iPacketSize--)
        pUdp->UDP_FDR[0] = *pData++ ;

      iCSR &= ~SET_FX;
      iCSR |= CLR_FX | AT91C_UDP_TXPKTRDY;

      pUdp->UDP_CSR[0] = iCSR;
//      pUdp->UDP_CSR[0] &= ~AT91C_UDP_TXPKTRDY;
//      pUdp->UDP_CSR[0] |= AT91C_UDP_TXPKTRDY;
      if (!(__sc_udp_waitTxCompleate(pUdp)))
        {
          UDP_DEBUG_WARNING("- write timeout stage 1\n");

          __sc_udp_handleTimeOut(pUdp);
          return;
        }
    } while (iLength);

    UDP_DEBUG_V3("- ok \n");
  }

/**
  Transmit an empty message.
*/
__inline void __sc_udp_sendZLP(AT91PS_UDP pUdp)
  {
    volatile unsigned int iCSR = 0;
    unsigned int iTimeOut = 1024*1024;

   // accept previous (failed) transmitions (paranoya) //
   pUdp->UDP_CSR[0] &= ~(AT91C_UDP_TXCOMP);

   if (!(__sc_udp_waitTxReady(pUdp)))
     {
       UDP_DEBUG_WARNING("\nUDP-SendZLP TimeOut Stage 0\n");
       __sc_udp_handleTimeOut(pUdp);
       return;
     }

    iTimeOut = 1024*1024;
    pUdp->UDP_CSR[0] |= AT91C_UDP_TXPKTRDY;
    if (!(__sc_udp_waitTxCompleate(pUdp)))
      {
        UDP_DEBUG_WARNING("\nUDP-ZLP TimeOut wait complete (CSR before: 0x"); scDbgPrintHex32(iCSR); scDbg(" | Now: 0x"); scDbgPrintHex32(pUdp->UDP_CSR[0]); scDbg("\n");
        __sc_udp_handleTimeOut(pUdp);
        return;
      }
  }

__inline void __sc_udp_sendStall(AT91PS_UDP pUdp)
  {
    unsigned int iTimeOut = 1024*1024;
    while ((pUdp->UDP_CSR[0] & AT91C_UDP_ISOERROR))
      {
        iTimeOut--;
        if (!(iTimeOut))
          {
            __sc_udp_handleTimeOut(pUdp);
            UDP_DEBUG_WARNING("\nUDP-STALL TimeOut stage 0\n");
            return;
          }
      }

    iTimeOut = 1024*1024;
    
    pUdp->UDP_CSR[0] |= AT91C_UDP_FORCESTALL;

    while (!(pUdp->UDP_CSR[0] & AT91C_UDP_ISOERROR))
      {
        iTimeOut--;
        if (!(iTimeOut))
          {
            __sc_udp_handleTimeOut(pUdp);
            UDP_DEBUG_WARNING("\nUDP-STALL timeout stage 1\n");
            return;
          }
      }
    pUdp->UDP_CSR[0] &= ~(AT91C_UDP_FORCESTALL | AT91C_UDP_ISOERROR);
  }
  
  
__inline void __sc_udp_readData(AT91PS_UDP pUdp, char* pBuffer, unsigned int iLength)
  {
    unsigned int iPacketSize = 0;
    unsigned int iTimeOut = 1024*1024;
    while (iLength)
      {
        iTimeOut = 1024*1024;
        while (!(pUdp->UDP_CSR[0] & AT91C_UDP_RX_DATA_BK0) && (--iTimeOut))
          ;;

        iPacketSize = MIN((pUdp->UDP_CSR[0] >> 16) & 0x7ff, iLength);
        iLength -= iPacketSize;
        if (iTimeOut != 0)
          {
            iTimeOut = 1024*1024;
            while ((pUdp->UDP_CSR[0] & AT91C_UDP_RX_DATA_BK0) && (--iTimeOut) && (iPacketSize))
              {
                *(pBuffer++) = pUdp->UDP_FDR[0];
                iPacketSize--;
              }
          }
          
        pUdp->UDP_CSR[0] &= ~(AT91C_UDP_RX_DATA_BK0);

        if (iTimeOut == 0)
          {
            UDP_DEBUG_WARNING("\nUDP-Read: read timeout\n");
            __sc_udp_handleTimeOut(pUdp);
            return;
          }
      }
  }


/**
  Handle usb messages.
  If there was an unknown message pOwnDispatcher will called.
  The return value of pOwnDispatcher will decide about the sending a usb-stall message.
  Return value of pOwnDispatcher:
    0 - message is unknown -> stall will be send
    1 - message was known  -> no stall will be send
*/
void __sc_udp_dispatchSetup(AT91PS_UDP pUDP, void* _pObject, char (*pOwnDispatcher)(void* _pObject, AT91PS_UDP pUDP, unsigned short iReq, unsigned short wValue, unsigned short wLength))
  {
    volatile unsigned char bmRequestType, bRequest, bStatus;
    volatile unsigned short wValue, wIndex, wLength, wStatus;
    volatile int iCSR = pUDP->UDP_CSR[0];

    if (!(iCSR & AT91C_UDP_RXSETUP))
        return;
      
    bmRequestType = pUDP->UDP_FDR[0];
    bRequest      = pUDP->UDP_FDR[0];
    wValue        = (pUDP->UDP_FDR[0] & 0xFF);
    wValue       |= (pUDP->UDP_FDR[0] << 8);
    wIndex        = (pUDP->UDP_FDR[0] & 0xFF);
    wIndex       |= (pUDP->UDP_FDR[0] << 8);
    wLength       = (pUDP->UDP_FDR[0] & 0xFF);
    wLength      |= (pUDP->UDP_FDR[0] << 8);

    if (bmRequestType & 0x80)
      pUDP->UDP_CSR[0] |= AT91C_UDP_DIR;
    pUDP->UDP_CSR[0] &= ~AT91C_UDP_RXSETUP;

    switch ((bRequest << 8) | bmRequestType)
      {
        case STD_GET_DESCRIPTOR:
          if (wValue == 0x100)       // Return Device Descriptor
            {
              UDP_DEBUG_V3("UDP get std desc.\n");
              __sc_udp_write(pUDP, devDescriptor, MIN(sizeof(devDescriptor), wLength));
            }
            else
            if (wValue == 0x200)  // Return Configuration Descriptor
              {
                UDP_DEBUG_V3("UDP get device desc.\n");
                __sc_udp_write(pUDP, cfgDescriptor, MIN(sizeof(cfgDescriptor), wLength));
              }
            else
            {
              UDP_DEBUG_V3("UDP unknown descriptor requested!\n");
                    __sc_udp_sendStall(pUDP);
            }
            break;
      
        case STD_SET_ADDRESS:
          UDP_DEBUG_V3("UDP-SetAddr: 0x"); UDP_DEBUG_V3_HEX(wValue); UDP_DEBUG_V3("\n");
          __sc_udp_sendZLP(pUDP);
          pUDP->UDP_FADDR = (AT91C_UDP_FEN | wValue);
          pUDP->UDP_GLBSTATE = (wValue) ? AT91C_UDP_FADDEN : 0;
         break;

        case STD_SET_CONFIGURATION:
          UDP_DEBUG_V3("UDP-SetConfig\n");
          giCurrentConfig = wValue;
          pUDP->UDP_GLBSTATE  = (wValue) ? AT91C_UDP_CONFG : AT91C_UDP_FADDEN;
          __sc_udp_sendZLP(pUDP);
         break;
          
        case STD_GET_CONFIGURATION:
          UDP_DEBUG_V3("UDP-GetCFG\n");
          __sc_udp_write(pUDP, (char*)&giCurrentConfig, sizeof(giCurrentConfig));
         return;

         
        case STD_GET_STATUS_ZERO:
          UDP_DEBUG_V3("UDP-GetStat-0\n");
        case STD_GET_STATUS_INTERFACE:
          UDP_DEBUG_V3("UDP-GetStat-I\n");
          wStatus = 0;
          __sc_udp_write(pUDP, (char *) &wStatus, sizeof(wStatus));
         return;

        case STD_SET_INTERFACE:
          // we only have one interface //
          __sc_udp_sendStall(pUDP);
         break;

        case STD_GET_STATUS_ENDPOINT:
          UDP_DEBUG_V3("UDP-GetStat-EP\n");
          bStatus = 0;
          if ((pUDP->UDP_GLBSTATE & AT91C_UDP_CONFG) && (wIndex == 0))
            {
              bStatus = (pUDP->UDP_CSR[wIndex] & AT91C_UDP_EPEDS) ? 1 : 0;
              __sc_udp_write(pUDP, (char*) &bStatus, sizeof(bStatus));
            }
          else
            if ((pUDP->UDP_GLBSTATE & AT91C_UDP_FADDEN) && (wIndex == 0))
              {
                bStatus = (pUDP->UDP_CSR[wIndex] & AT91C_UDP_EPEDS) ? 1 : 0;
                __sc_udp_write(pUDP, (char*)&bStatus, sizeof(bStatus));
              }
            else
              __sc_udp_sendStall(pUDP);
         break;

        case STD_SET_FEATURE_ZERO:
          UDP_DEBUG_V3("UDP-SFeatureZero\n");
          __sc_udp_sendStall(pUDP);
         break;

        case STD_SET_FEATURE_INTERFACE:
          UDP_DEBUG_V3("UDP-SFeatureIface\n");
          __sc_udp_sendZLP(pUDP);
         break;

        case STD_SET_FEATURE_ENDPOINT:
          UDP_DEBUG_V3("UDP-SFeatureEP\n");
          __sc_udp_sendStall(pUDP);
         break;
         
        case STD_CLEAR_FEATURE_ZERO:
          __sc_udp_sendStall(pUDP);
         break;
        case STD_CLEAR_FEATURE_INTERFACE:
          UDP_DEBUG_V3("UDP-CFeatureFI\n");
          __sc_udp_sendZLP(pUDP);
         break;
        case STD_CLEAR_FEATURE_ENDPOINT:
          UDP_DEBUG_V3("UDP-CFeatureFEP\n");
          __sc_udp_sendStall(pUDP);
         break;

         
        default:
          {
            char boKnown = 0;
            if (pOwnDispatcher)
              boKnown = pOwnDispatcher(_pObject, pUDP, (bRequest << 8) | bmRequestType, wValue, wLength);

            if (!(boKnown))
              {
                UDP_DEBUG_WARNING("unknown request: 0x"); UDP_DEBUG_WARNING_HEX((bRequest << 8) | bmRequestType); UDP_DEBUG_WARNING("\n");
                __sc_udp_sendStall(pUDP);
              }
          }
      }
  }




void sc_udp_init()
  {
    UDP_DEBUG_V3("\ninit udp...");
    // Configure PMC
//////    AT91F_UDP_CfgPMC();                        // Enable UDP peripheral
    AT91C_BASE_PMC->PMC_PCER = (unsigned int)1 << AT91C_ID_UDP;
//////
    AT91C_BASE_PMC->PMC_SCER |= AT91C_PMC_UDP; // Enable 48 MHz
    giCurrentConfig = 0;
    // setup pio port PD9 //
    AT91C_BASE_PIOD->PIO_SODR = AT91C_PIO_PD9;
//    AT91C_BASE_PIOD->PIO_PER = AT91C_PIO_PD9;
    AT91C_BASE_PIOD->PIO_OER = AT91C_PIO_PD9;
  }

void sc_udp_resetBUS(void)
  {
    unsigned int i;
    UDP_DEBUG_V3("UDP-Resetting bus... \n");
    AT91C_BASE_PIOD->PIO_OER = SC_IO_CONFIG_UDP_DISCONNECT_BUS;
    AT91C_BASE_PIOD->PIO_CODR = SC_IO_CONFIG_UDP_DISCONNECT_BUS;
    for (i = 0; i < 10000; i++);
    AT91C_BASE_PIOD->PIO_SODR = SC_IO_CONFIG_UDP_DISCONNECT_BUS;
    // We need this to allow the manual reset //
    AT91C_BASE_PIOD->PIO_PPUER = SC_IO_CONFIG_UDP_DISCONNECT_BUS;
    AT91C_BASE_PIOD->PIO_ODR = SC_IO_CONFIG_UDP_DISCONNECT_BUS;
  }

void sc_udp_done()
  {
    UDP_DEBUG_V3("\nshutdown udp...");
    // Configure PMC
    AT91C_BASE_PMC->PMC_SCDR |= AT91C_PMC_UDP; // Disable 48 MHz
    giCurrentConfig = 0;
    sc_udp_resetBUS();
  }


void sc_udp_poll(void* _pObject, char (*pOwnDispatcher)(void* _pObject, AT91PS_UDP pUDP, unsigned short iReq, unsigned short wValue, unsigned short wLength))
  {
    if (AT91C_BASE_UDP->UDP_ISR & (1 << 13))
      AT91C_BASE_UDP->UDP_ICR |= (1 << 13);
    
    if (AT91C_BASE_UDP->UDP_ISR & (1 << 12))
      {
        UDP_DEBUG_V3("\nbus reset finished\n");
        AT91C_BASE_UDP->UDP_IDR = 0xFFFFFFFF ;
        // Disable all pending IT
        AT91C_BASE_UDP->UDP_ICR = 0xFFFFFFFF ;
        // RESET UDP
        AT91C_BASE_UDP->UDP_RSTEP  = 0;
        AT91C_BASE_UDP->UDP_GLBSTATE = 0;
        AT91C_BASE_UDP->UDP_ICR = (1 << 12);
        AT91C_BASE_UDP->UDP_CSR[0] |= AT91C_UDP_EPTYPE_CTRL | AT91C_UDP_EPEDS;
      }
    else
      {
        __sc_udp_dispatchSetup(AT91C_BASE_UDP, _pObject, pOwnDispatcher);
      }
  }

