/**
 * Copyright (c) 2005 Peer Georgi <peer@peer-ware.de>
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for complete details.
*/


#include "cNetAbstractProtocol.h"
#include "cNetProtocolMgr.h"
#include "cAbstractNetDevice.h"

/**
  Use this constructor to create a new sub-protocol.
*/
cNetAbstractProtocol::cNetAbstractProtocol(cNetAbstractProtocol& parentProtocol)
  : mNetProtocolMgrNode(this), mNetSubProtocolNode(this), mpMgrParent(0), mpProtocolParent(&parentProtocol)
  {
    parentProtocol.registerSubProtocol(*this);
  }

/**
  Use this constructor to create a new top-level-protocol (family).
*/
cNetAbstractProtocol::cNetAbstractProtocol(cNetProtocolMgr& parent)
  : mNetProtocolMgrNode(this), mNetSubProtocolNode(this), mpMgrParent(&parent), mpProtocolParent(0)
  {
    parent.registerProtocol(*this);
  }


cNetAbstractProtocol::~cNetAbstractProtocol() throw()
  {
  }


cAbstractNetDevice& cNetAbstractProtocol::getDevice() const
  {
    if (mpProtocolParent)
      return (mpProtocolParent->getDevice());
    return (mpMgrParent->getDevice());
  }

cNetProtocolMgr& cNetAbstractProtocol::getMgr() const
  {
    if (mpMgrParent)
      return (*mpMgrParent);
    return (mpProtocolParent->getMgr());
  }

/**
  This is for dispatching a received and verified frame to sub-layers of this protocol.
  The method may be called in handleFrame(..) on successfully read a sublayer frame.
  The parameters are pFrame and iFrameSize for the size of unwrapped frame.
  pFull und iFullSize should contain the original unwrapped frame (the parameters of handleFrame(..))
*/
cNetAbstractProtocol::RxHandleState cNetAbstractProtocol::dispatchFrameToSubProtocol(char* pFrame, unsigned int iFrameSize, char* pFull, unsigned int iFullSize)
  {
    cNetAbstractProtocol::RxHandleState iResult = cNetAbstractProtocol::rxhNotSupported;
    for (c_cSLList<cNetAbstractProtocol>::iterator aIt = mLstSubProtocols.begin(); aIt != mLstSubProtocols.end(); ++aIt)
      {
        iResult = (*aIt)->handleFrame(pFrame, iFrameSize, pFull, iFullSize);
        if (iResult != cNetAbstractProtocol::rxhNotHandled)
          break;
      }
    return (iResult);
  }

/**
*/
void cNetAbstractProtocol::pushPacket(char* pPacket, unsigned int iSize, const cNetAbstractPEParam* pParam)
  {
    if (mpProtocolParent)
      mpProtocolParent->encapsulatePacket(pPacket, iSize, pParam);
    else
      if (mpMgrParent)
        mpMgrParent->getDevice().send(pPacket, iSize);
  }

extern "C"
  {
    #include <sc/sc.h>
  }


void cNetAbstractProtocol::dispatchProtocolParamRecursive(const cNetAbstractProtocolParam& p)
  {
    notifyProtocolParam(p);
    for (c_cSLList<cNetAbstractProtocol>::iterator aIt = mLstSubProtocols.begin(); aIt != mLstSubProtocols.end(); ++aIt)
      (*aIt)->dispatchProtocolParamRecursive(p);
  }

const cNetAbstractProtocolParam* cNetAbstractProtocol::findProtocolParamRecursive(unsigned int iID)
  {
    const cNetAbstractProtocolParam* p = hasProtocolParam(iID);
    if (p) return (p);
    for (c_cSLList<cNetAbstractProtocol>::iterator aIt = mLstSubProtocols.begin(); aIt != mLstSubProtocols.end(); ++aIt)
      {
        p = (*aIt)->findProtocolParamRecursive(iID);
        if (p) return (p);
      }
    return (0);
  }


void cNetAbstractProtocol::registerSubProtocol(cNetAbstractProtocol& sp)
  {
    mLstSubProtocols.append(&sp.mNetSubProtocolNode);
  }

/// To Implement!
cNetAbstractProtocol::RxHandleState cNetAbstractProtocol::handleFrame(char*, unsigned int, char*, unsigned int)
  {
    return (rxhNotHandled);
  }

/// To Implement!
void cNetAbstractProtocol::notifyProtocolParam(const cNetAbstractProtocolParam&)
  {
  }

/// To Implement!
const cNetAbstractProtocolParam* cNetAbstractProtocol::hasProtocolParam(unsigned int iID)
  {
    return (0);
  }

/**
  This method may called by an higher protocol-layer.
*/
void cNetAbstractProtocol::encapsulatePacket(char* pPacket, unsigned int iSize, const cNetAbstractPEParam* pParam)
  {
  }
