/**
 * Copyright (c) 2005 Peer Georgi <peer@peer-ware.de>
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for complete details.
*/


#include "cNetProtoARP.h"
#include "cNetMACProtocolParam.h"
#include "cAbstractNetDevice.h"
#include "cNetProtocolID.h"
#include "cNetMACProtocolPEParam.h"
#include "cNetARPProtocolPEParam.h"
#include "cNetProtocolMgr.h"
#include "tinylibc/string.h"
#include "console.h"

extern "C"
  {
    #include <sc/sc.h>
  }


cNetProtoARP::cNetProtoARP(cNetAbstractProtocol& parent)
  : cNetAbstractProtocol(parent), mProtocolParam()
    {
    }

cNetProtoARP::~cNetProtoARP() throw()
  {
  }

void cNetProtoARP::notifyProtocolParam(const cNetAbstractProtocolParam& prm)
  {
    if (prm.getID() == mProtocolParam.getID())
      {
        cNetIPv4ProtocolParam& IPPrm = (cNetIPv4ProtocolParam&)prm;
        mProtocolParam = IPPrm;
      }
  }


const cNetAbstractProtocolParam* cNetProtoARP::hasProtocolParam(unsigned int iID)
  {
    if (iID == mProtocolParam.getID())
      return (&mProtocolParam);
    return (0);
  }


struct cNetProtoARP::tMACFrame
  {
    char pDest[6];
    char pFrom[6];
    char pPad[0x600-12];
  };

struct cNetProtoARP::tARPFrame
  {
    unsigned short iProtocolLayer;
    unsigned short iHWType;
    unsigned short iProtocolType;
    char iHWSize;         // must be 6
    char iProtocolSize;   // must be 4
    unsigned short iOpcode;
    char pSenderMAC[6];
    char pSenderIP[4];
    char pTargetMAC[6];
    char pTargetIP[4];
    char pTraylor[0x600]; // TODO Wrong value
  } __attribute__((packed));


cNetAbstractProtocol::RxHandleState cNetProtoARP::handleFrame(char* pFrame, unsigned int iSize, char* pFullFrame, unsigned int)
  {
    cNetProtoARP::tARPFrame* p = (tARPFrame*)pFrame;
    if (p->iProtocolLayer != 0x0608) //  ! ARP //
      {
        if (p->iProtocolLayer == 0x0008) // IP //
          return (dispatchFrameToSubProtocol(pFrame, iSize, pFrame, iSize));
        return (cNetAbstractProtocol::rxhNotHandled);
      }

    if (iSize < 30) return (cNetAbstractProtocol::rxhNotHandled);

    if ((p->iHWSize != 6) || (p->iProtocolSize != 4))
      {
        scDbg("ARP: Ipv6 not supported yet!\n");
        return (cNetAbstractProtocol::rxhNotSupported);
      }

    if (p->iProtocolType != 0x0008) // ! IP //
      {
        scDbg("ARP: unsupprted protocol type!\n");
        return (cNetAbstractProtocol::rxhNotSupported);
      }
    switch (p->iOpcode)
      {
      case 0x0100:
         handleWhoHasIPv4(p, iSize);
        break;
      default:
        scDbg("ARP: unsupported opcode (0x"); scDbgPrintHex8(p->iOpcode >> 8);  scDbgPrintHex8(p->iOpcode & 255); scDbg(")!\n");
        return (cNetAbstractProtocol::rxhNotSupported);
      };
    return (cNetAbstractProtocol::rxhHandled);    
  }

// Valid size of p is at least 30 bytes!
void cNetProtoARP::handleWhoHasIPv4(tARPFrame* p, unsigned int iSize)
  {
    cNetIPv4Address ipAddrSource;
    cNetIPv4Address ipAddrTarget;
    cNetMACAddress macAddrSource;
    ipAddrSource.load(p->pSenderIP);
    ipAddrTarget.load(p->pTargetIP);
    macAddrSource.load(p->pSenderMAC, 6);
    
        
        scDbg("ARP: Who has ");
        ipAddrTarget.debugPrint();
        scDbg(" from ");
        ipAddrSource.debugPrint();
        scDbg(" ?\n");
    // lead ARP cache //
    mARPCache.addCoherence(ipAddrSource, macAddrSource);

    if (ipAddrTarget == mProtocolParam.getAddress())
      { // answer frame //
        
        memcpy(p->pTargetIP, p->pSenderIP, 6);
        memcpy(p->pTargetMAC, p->pSenderMAC, 6);
        cNetMACProtocolParam* pMACProtocolParam = (cNetMACProtocolParam*)getMgr().getProtocolParam(NET_PROTOCOL_ID_MAC);
        if (!(pMACProtocolParam))
          {
            scDbg("MAC layer not found!");
            return;
          }
        pMACProtocolParam->getAddress().store(p->pSenderMAC);
        mProtocolParam.getAddress().store(p->pSenderIP);
        p->iOpcode = 0x0200;

        cNetMACProtocolPEParam prm(cNetMACAddress(p->pTargetMAC, 6));
        pushPacket((char*)p, 30, &prm); // PADDING will be added by mac automaticaly //
      }
  }

/**
  We have to look for the MAC address in ARP cache.
*/
void cNetProtoARP::encapsulatePacket(char* pPacket, unsigned int iSize, const cNetAbstractPEParam* pParam)
  {
    if (!(pParam)) return;
    if (pParam->getID() != NET_PROTOCOL_ID_ARP) return;
    cNetARPProtocolPEParam* pIPProtocolParam = (cNetARPProtocolPEParam*)pParam;
    const cNetMACAddress* pMACAddress = mARPCache.searchMACAddress(pIPProtocolParam->getAddress());
    if (pMACAddress)
      {
        scDbg("ARP: encap packet of size "); scDbgPrintDec(iSize); scDbg("\n");
        cNetMACProtocolPEParam macPrm(*pMACAddress);
        pushPacket((char*)pPacket, iSize, &macPrm);
      }
    else
      {
        // The following is todo:
        //   * we have to push the given packet to a queue //
        //   * ask for ip-address (arp-request)
        //   * the response to arp-request must ivoke to pop the queued packet and immediately sending
        scDbg("ARP: Unable to resolve "); pIPProtocolParam->getAddress().debugPrint(); scDbg(" ARP request not supported yet.\n");
      }
//    cNetMACProtocolPEParam prm(pTargetMACAddress, 6);
  }

