/**
 * Copyright (c) 2005 Peer Georgi <peer@peer-ware.de>
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for complete details.
*/


#include "cNetProtoIPv4.h"
#include "cNetAbstractPEParam.h"
#include "cNetIPv4ProtocolPEParam.h"
#include "cNetARPProtocolPEParam.h"
#include "cNetProtocolID.h"
#include "cNetProtocolMgr.h"
#include "console.h"
#include <tinylibc/string.h>

extern "C"
  {
    #include <sc/sc.h>
  }


cNetProtoIPv4::cNetProtoIPv4(cNetAbstractProtocol& parent)
  : cNetAbstractProtocol(parent), mProtocolParam()
    {
    }

cNetProtoIPv4::~cNetProtoIPv4() throw()
  {
  }

void cNetProtoIPv4::notifyProtocolParam(const cNetAbstractProtocolParam& prm)
  {
    if (prm.getID() == mProtocolParam.getID())
      {
        cNetIPv4ProtocolParam& IPPrm = (cNetIPv4ProtocolParam&)prm;
        mProtocolParam = IPPrm;
      }
//    scDbg("IP: using address: "); mProtocolParam.getAddress().debugPrint(); scDbg("\n");
  }


const cNetAbstractProtocolParam* cNetProtoIPv4::hasProtocolParam(unsigned int iID)
  {
    if (iID == mProtocolParam.getID())
      return (&mProtocolParam);
    return (0);
  }


struct tIPv4Frame
  {
    unsigned short iProtocolLayer;
    unsigned char bIDAndHeaderLength;
    unsigned char bECN;
    unsigned short iTotalLength;
    unsigned short iIdentification;
    unsigned char bFlags;
    unsigned char bFrameOffset;
    unsigned char bTTL;
    unsigned char bIPProto; // 1 ICMP
    unsigned short iHeaderCheckSum;
    char pSourceIP[4];
    char pDestIP[4];
    char bType;             // 8 echo ping
    char bCode;             // 0
    char pData[2048];
  } __attribute__((packed));


cNetAbstractProtocol::RxHandleState cNetProtoIPv4::handleFrame(char* pFrame, unsigned int iSize, char* pFullFrame, unsigned int iFullFrameSize)
  {
    if (iSize < 24) return (rxhNotHandled);
    tIPv4Frame* pIPv4Frame = (tIPv4Frame*)pFrame;
    scDbg("IP: frame received "); scDbgPrintDec(iSize); scDbg(" bytes.\n");
    cNetIPv4Address addrTarget;
    addrTarget.load(pIPv4Frame->pDestIP);
    if (addrTarget != mProtocolParam.getAddress())
      return (rxhNotHandled);

    if ((pIPv4Frame->bFrameOffset) || (pIPv4Frame->bFlags & 2))
      {
        scDbg("IP: Multible fragments are not supported yet (sorry)\n");
        return (cNetAbstractProtocol::rxhNotSupported);
      }
    
    // Check Checksum
    unsigned int iByteLen = (pIPv4Frame->bIDAndHeaderLength & 0xf) << 2; // length is given in 32-bit words //

    unsigned short iOldSum = pIPv4Frame->iHeaderCheckSum; // to compare it later //
    unsigned short iNewSum = 0;

    if (iOldSum != 0)
      {
        pIPv4Frame->iHeaderCheckSum = 0;                   // precondition for checksum calculation //
        unsigned short* pIPFrame = (unsigned short*)&pIPv4Frame->bIDAndHeaderLength; // point to the ip-header //
        unsigned int iWSum = 0;                   // 32-bit sum //
        for (unsigned int i = 0; i < iByteLen >> 1; i++)
          iWSum += pIPFrame[i];
        iNewSum = (iWSum >> 16)+(iWSum & 0xffff); // 16-bit sum //
        iNewSum = ~iNewSum;                       // complement of two //
      }
    if (iOldSum != iNewSum)
      return (rxhHandled); // drop errornous frames //

//    scDbg(" Checksum: 0x"); scDbgPrintHex32(iNewSum);     
   
    // TODO: Broadcast detection (http://www.erg.abdn.ac.uk/users/gorry/course/inet-pages/ip-processing.html)
    char* pFrameToSub = pFrame+14;
    unsigned int iSubSize = iSize-14;
    return (dispatchFrameToSubProtocol(pFrameToSub, iSubSize, pFullFrame, iFullFrameSize));
  }


void cNetProtoIPv4::encapsulatePacket(char* pPacket, unsigned int iSize, const cNetAbstractPEParam* pParam)
  {
    if (!(pParam)) return;
    if (!(pParam->getID() == NET_PROTOCOL_ID_IPV4)) return;
//    scDbg("IP: encap packet of size "); scDbgPrintDec(iSize); scDbg("\n");
    cNetIPv4ProtocolPEParam* pIPv4Prm = (cNetIPv4ProtocolPEParam*)pParam;
    tIPv4Frame frame;
    frame.iProtocolLayer = 0x0008;
    frame.bIDAndHeaderLength = 0x45;
    frame.bECN = 0;
    frame.iTotalLength = (iSize+12) << 8;
    frame.iIdentification = 0; // TODO: Ident-Nr-
    frame.bFlags = 0;          // TODO: Information about fragments
    frame.bFrameOffset = 0;
    frame.bTTL = 64;
    frame.bIPProto = pIPv4Prm->getProtocol();
    frame.iHeaderCheckSum = 0;

    // Check Checksum
    unsigned int iByteLen = (frame.bIDAndHeaderLength & 0xf) << 2; // length is given in 32-bit words //
    unsigned short iNewSum = 0;
    unsigned short* pIPFrame = (unsigned short*)&frame.bIDAndHeaderLength; // point to the ip-header //
    unsigned int iWSum = 0;                   // 32-bit sum //
    for (unsigned int i = 0; i < iByteLen >> 1; i++)
      iWSum += pIPFrame[i];
    iNewSum = (iWSum >> 16)+(iWSum & 0xffff); // 16-bit sum //
    iNewSum = ~iNewSum;                       // complement of two //

    frame.iHeaderCheckSum = iNewSum;
    memcpy(&frame.bType, pPacket+8, iSize-8);
    mProtocolParam.getAddress().store(frame.pSourceIP);
    pIPv4Prm->getAddress().store(frame.pDestIP);
    cNetARPProtocolPEParam arpPrm;
    arpPrm.setAddress(pIPv4Prm->getAddress()); // dest ip as parameter for arp //
    pushPacket((char*)&frame, iSize+26-8, &arpPrm);
   }

