/**
 * Copyright (c) 2005 Peer Georgi <peer@peer-ware.de>
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for complete details.
*/


#ifndef __c_tSLList_h
  #define __c_tSLList_h

  
/*
  Hier werden Templates geboten, welche eine einfache Implementierung von einfach verketteten Listen
  erlauben.
*/



template <class cNodeType> class c_cSLList
  {
  public:
/**
  Ein spezialisiertes Objekt dieses Templates wird in der Klasse als Attribut verankert, welche
  gelistet werden soll.
*/
class cNode
  {
  class iterator;

  public:
    cNode(cNodeType* pCarrier = 0)
      {
        mpList = 0;
        mpNext = 0;
        mpNodeCarrier = pCarrier;
      }

    ~cNode()
      {
        if (mpList)
          mpList->removeNode(this);
      }

    void setCarrier(cNodeType* p)
      {
        mpNodeCarrier = p;
      }

    bool operator==(cNode& a)
      {
        if (!(mpList == a.mpList)) return (false);
        if (!(mpNext == a.mpNext)) return (false);
        if (!(mpNodeCarrier == a.mpNodeCarrier)) return (false);
        return (true);
      }

    bool operator!=(cNode& a)
      {
        return (!(*this == a));
      }

    cNode& operator++()
      {
        if (!(mpList)) return (*this);
        if (!(mpNext)) return (mpList->mEndNode);
        return (*mpNext);
      }

    void unlink()
      {
        mpList = 0;
        mpNext = 0;
      }

    c_cSLList<cNodeType>* mpList;
    cNode* mpNext;
    cNodeType* mpNodeCarrier; // Die Klasse, welches diese Struktur enthlt //
  };

  public:

    /**
      Use an object of this class to iterate the list.
      You can use operator "++" (prefix) to increment the position. The maximum position is
      c_cSLList::end(). Also you can derefer any members of the class you lead in the list with the
      operator "->". The operator "*" (prefix) is overloaded too to allow easy access to your object.
    */
    class iterator
      {
      public:
        iterator()
          {
            mpNode = 0;
          }
        iterator(cNode& aEn)
          {
            mpNode = &aEn;
          }
        iterator& operator=(cNode& aEn)
          {
            mpNode = &aEn;
            return (*this);
          }
        bool operator==(iterator& a)
          {
            return (mpNode == a.mpNode);
          }
        bool operator!=(iterator& a)
          {
            return (mpNode != a.mpNode);
          }
        bool operator==(cNode& a)
          {
            if (!(mpNode)) return (false);
            return (*mpNode == a);
          }
        bool operator!=(cNode& a)
          {
            if (!(mpNode)) return (false);
            return (*mpNode != a);
          }
        cNodeType* operator*()
          {
            if (!(mpNode)) return (0);
            return (mpNode->mpNodeCarrier);
          }
        cNodeType* operator->()
          {
            if (!(mpNode)) return (0);
            return (mpNode->mpNodeCarrier);
          }
        iterator& operator++()
          {
            if (!(mpNode)) return (*this);
            if (mpNode->mpNext)
              mpNode = mpNode->mpNext;
            return (*this);
          }
        iterator& operator++(int)
          {
            return (++(*this));
          }
      private:
        cNode* mpNode;
      };

  /**
    Single linked list.
  */
  public:

    /**
      Creates a new empty list.
    */
    c_cSLList() : mEndNode(0)
      {
        mpRoot = &mEndNode;
      }

    /**
      Destroys the list structure.
      All nodes will be removed using clear().
    */
    ~c_cSLList()
      {
        clear();
      }

    /**
      Returns the first element in list.
      If the list is empty the stop-element will be returned.
    */
    cNode& begin()
      {
        return (*mpRoot);
      }

    /**
      Returns the stop element.
      The stop-element does not contain a valid referance to
      an object. It leads an undefined (zero) mpNodeCarrier.
    */
    cNode& end()
      {
        return (mEndNode);
      }

    /**
      Returns true if the list is empty.
    */
    bool empty()
      {
        return (mpRoot == &mEndNode);
      }

    /**
      Append the list with a new node.
      We lead a stop-element. That means all append
      operations are equal with an insert before the end-node.
    */
    void append(cNode* pNode)
      {
        if (!(pNode)) return;
        pNode->mpList = this;
        pNode->mpNext = &mEndNode;
        cNode* pEn = mpRoot;
        cNode* pLast = 0;
        while (pEn != &mEndNode)
          {
            pLast = pEn;
            pEn = pEn->mpNext;
          }
        if (!(pLast))
          mpRoot = pNode;
        else
          pLast->mpNext = pNode;
      }

    /**
      Remove a list node. This takes care about the
      node- and list consistency. The destructor of
      a node calls this method automaticaly.
    */
    void removeNode(cNode* pNode)
      {
        if (!(pNode)) return;
        // wir suchen das Element //
        cNode* pEn = mpRoot;
        cNode* pLast = 0;
        while (pEn)
          {
            if (pEn == pNode)
              break;
            pLast = pEn;
            pEn = pEn->mpNext;
          }
        if (!(pEn)) return; // not in list //
        if (!(pLast)) // no element before -> replace root-element //
          mpRoot = pEn->mpNext;
        else
          pLast->mpNext = pEn->mpNext;
        pNode->unlink();
      }

    /**
      Clean ups the list.
      All nodes will be removed using removeNode(..) except the
      stop element.
    */
    void clear()
      {
        while ((mpRoot) && (mpRoot != &mEndNode))
          removeNode(mpRoot);
      }

  private:
    cNode* mpRoot;
    cNode mEndNode;
  };

#endif // __c_tSLList_h //
