/**
 *
 * Copyright (c) 2005,2006 Conitec Datasystems GmbH
 * Written 2005,2006 by Peer Georgi <georgi@conitec.com>
 *
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for complete details.
 **
*/


/**
  RELOCATION: 0x2007a000

  This application writes data into the dataflash device.
  Theese data must be stored in SDRAM before.
  Additionaly you have to write a setup-list to the first address of sdram.
  The structure of this list is descripted below. (use ebFlashSetup to generate all structures)
*/

#include <sc/sc.h> // system core //
#include <sc/sc_config.h>
#include <sc/sys_core.h>
#include <libdf/spi_dataflash.h>
typedef int bool;
#define true  (1 == 1)
#define false (!true)

extern void ledSetSingle(unsigned int iIdx);
extern void ledToggle(void);
extern void ledInit(void);



//Memory Layout:
//  0x20070000 - setup entry list (max. 64kB)
//               with items of the following structure:
//                 * Entry-Name:         char[13]     (inclusive zero termination)
//                 * Entry-DFAddress:    UInt32         (byte address where the object is to store in dataflash)
//                 * Entry-LoadAddress:  UInt32       (byte address where the object is loaded in memory while setup procedure)
//                 * Entry-Size:         UInt32       (size of object in bytes)
//
//              The first item must be "setup-list" with all fields = 0 //
//
//              The last item in list must show: Entry-Name[0] == 0.
//
//              All multi-byte types are little-endian types.
//
//  0x2007a000 - this application (max. 512k)  (+0,5Mb)
//


struct cItemEntry
  {
    char msItemName[16];            // zero terminated msItemName[0] == 0 means end of list //
    unsigned int miDFStartAddress;  // start address in dataflash //
    unsigned int miLoadAddress;     // address where the data copied from //
    unsigned int miSize;            // size of data object //
  };

int strncmp(const char * cs, const char * ct,unsigned int count)
  {
    register signed char bRes = 0;
    while (count)
      {
        if ((bRes = *cs - *ct++) != 0 || !*cs++)
          break;
        count--;
      }
    return bRes;
  }

char* memcpy(char* pDst, const char* pSrc, unsigned int iSize)
  {
    register unsigned int i = 0;
    for (i = 0; i < iSize; i++)
      pDst[i] = pSrc[i];
    return (pDst);
  }

bool compareBuffers(const char* p1, const char* p2, unsigned int iSize)
  {
    register unsigned int i = 0;
    for (i = 0; i < iSize; i++)
      if (p1[i] != p2[i])
        return (false);
    return (true);
  }

struct tDFDevice gBootFlash;

bool dfWriteAndVerify(unsigned int iDFAddr, unsigned int iLoadAddress, unsigned int iSize)
  {
    static char pPreBuffer[21120];  // must be static due to limited stack size //
    static char pPostBuffer[21120]; // must be the same size //
    unsigned int iBufferSize = sizeof(pPreBuffer);
    char* pSource = (char*)iLoadAddress;
    unsigned int iBytesLeft = iSize;

    while (iBytesLeft)
      {
        unsigned int iBytesToWrite = iBytesLeft;
        if (iBytesToWrite > iBufferSize)
          iBytesToWrite = iBufferSize;
        memcpy(pPreBuffer, pSource, iBytesToWrite);
        // write data //
        // [
        if (spiDFWrite(&gBootFlash, pSource, iDFAddr, iBytesToWrite) < 0)
          {
            scDbg("\nDataflash write error!\n");
            return (false);
          }
        scDbg(".");
        ledToggle();
        if (!(spiDFRead(&gBootFlash, iDFAddr, pPostBuffer, iBytesToWrite)))
          {
            scDbg("\nVerify problem (unable to read)\n");
            return (false);
          }

        // ]
        // verify //
        // [
        if (!(compareBuffers(pPreBuffer, pPostBuffer, iBytesToWrite)))
          {
            scDbg("\nVerify failed (data not equal)\n");
            return (false);
          }

        // ]
        iBytesLeft -= iBytesToWrite;
        pSource += iBytesToWrite;
        iDFAddr += iBytesToWrite;
      }
    return (true);
  }

bool dfSetBootableFlag(bool bo, unsigned int iItemSize)
  {
    char pBuffer[512];
    if (!(spiDFRead(&gBootFlash, 0, pBuffer, 512)))
      {
        scDbg("\nProblem while setting bootable flag (unable to read)\n");
        return (false);
      }

    if (bo)
      { // set bootable flag //
        int j = 0;
        int i = 0;
        int iPageSize = gBootFlash.miNumPages;
        while (iPageSize >>= 1)
          j++;
        i = (iItemSize / 512) + 1 + (j << 13) + (gBootFlash.miPageSize << 17);
        *(int*)(pBuffer+0x14) = i;
      }
    else
      { // remove bootloader //
        unsigned int i = 0;
        for (i = 0; i < 512; i++)
          pBuffer[i] = 0;
      }
    return (dfWriteAndVerify(0, (unsigned int)pBuffer, 512));
  }

void boot(void)
  {
    struct cItemEntry* pCurrentItem = (struct cItemEntry*)0x20070000;
    bool boMakeBootable = false;
    unsigned int iBootableItemSize = 0;
    sc_sti();
    sc_waitTime(200);
    scDbg("\nDataFlash setup v1.3 (c) 2004, 2005, 2006 by Conitec Datensysteme GmbH.\n\n");

    spiDFInit(60000000, 15000000);
    spiDFProbeDevice(0, &gBootFlash);

    ledInit();

    // check signature //
    if (strncmp(pCurrentItem->msItemName, "setup-list", 10))
      {
        scDbg("You have to transfer a valid setup list before!\n");
        scDbg("(giving up)\n");
      }
    else
      {
        pCurrentItem++; // skip signature item //
        if (!(gBootFlash.miNumPages))
          {
            scDbg("Error initializing dataflash!\n");
            scDbg("Please check hardware...\n");
          }
        else
          {
            scDbg("Dataflash found: "); scDbgPrintHex32(gBootFlash.miNumPages); scDbg(" pages\n\n");
            while (1)
              {
                if (pCurrentItem->msItemName[0] == 0)
                  break; // last item //
                scDbg("Setting up \""); scDbg(pCurrentItem->msItemName); scDbg("\" at 0x"); scDbgPrintHex32(pCurrentItem->miLoadAddress);
                scDbg(" of size 0x"); scDbgPrintHex32(pCurrentItem->miSize); scDbg("...\n");
                scDbg(" * writing to dataflash address 0x"); scDbgPrintHex32(pCurrentItem->miDFStartAddress);

                if (pCurrentItem->miDFStartAddress == 0)
                  {
                    boMakeBootable = true;
                    iBootableItemSize = pCurrentItem->miSize;
                    scDbg(" (make bootable later) ");
                  }

                scDbg("...");
                if (!(dfWriteAndVerify(pCurrentItem->miDFStartAddress, pCurrentItem->miLoadAddress, pCurrentItem->miSize)))
                  {
                    scDbg("\n\n --- Flash setup failed ! --- \n\n");
                    // remove bootloader //
                    // [
                      boMakeBootable = false;
                      iBootableItemSize = 0;
                    // ]
                    break;
                  }
                scDbg(" - ok\n");
                scDbg("\n");
                pCurrentItem++;
              }
          }
      }

    scDbg((boMakeBootable) ? ("Setting") : ("Removing"));
    scDbg(" bootable flag...");
    if (dfSetBootableFlag(boMakeBootable, iBootableItemSize))
      scDbg (" - done.\n");
    else
      scDbg (" - failed!\n");

    scDbg("\n\n *** please restart target... ***");

    {
      int iRun = 0;
      int iLEDIdx = 0;
      int iLEDDir = 1;
      while (1)
        {
          sc_waitTime(50);
          if (iLEDIdx >= 3)
            {
              iLEDDir *= -1;
              iLEDIdx = 3;
              iRun++;
            }
          if (iLEDIdx < 0)
            {
              iLEDIdx = 1;
              iLEDDir *= -1;
            }
          ledSetSingle(iLEDIdx);
          iLEDIdx += iLEDDir;

        if (iRun > 10)
          {
            scDbg("\nWill restart now...\n");
            scReset();
          }
        }
     }
  }
