/**
 * Copyright (c) 2005 Peer Georgi <peer@peer-ware.de>
 * License: GPL
 *
 * This program is free software; you may redistribute and/or modify it under
 * the terms of the GNU General Public License Version 2 as published by the
 * Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for complete details.
*/


/**
  RELOCATION: 0x21000000.
*/
extern "C"
  {
    #include <sc/sc.h> // system core //
    #include <sc/sc_config.h>
    #include <at91rm9200-def.h>
  }

#include "cPin.h"
#include "console.h"

extern "C" void boot();


void showHelp()
  {
    scDbg("\ncArmEvaIOTest (c) 2005 Conitec Datasystems GmbH.\n\n");
    scDbg(" \"w\" - check with jumper on J8, J10 and J11\n");
    scDbg(" \"W\" - check without jumoer (but pass them)\n");
    scDbg(" \"J\" - check without jumoer (stricly)\n");
    scDbg("\n");
  }

class cTestBridge
  {
  public:
    cTestBridge(unsigned int iIndex, const char* sConnector, cPin pin1, cPin pin2)
      : miIndex(iIndex), mPin1(pin1), mPin2(pin2)
        {
          mPin1.setComment(sConnector);
          mPin2.setComment(sConnector);
        }

    unsigned int index() const
      { return (miIndex); }

    /// \brief Setup all pins to low.
    void clrAll()
      {
        mPin1.clr();
        mPin2.clr();
      }

    /// \brief Setup all pins to low.
    void setAll()
      {
        mPin1.set();
        mPin2.set();
      }

    /// \brief Returns the pin which indicates (reads) high or zero if booth are low.
    cPin* hasHigh()
      {
        if (mPin1.get())
          return (&mPin1);
        if (mPin2.get())
          return (&mPin2);
        return (0);
      }
    /// \brief Returns the pin which indicates (reads) low. Returns zero if booth are high.
    cPin* hasLow()
      {
        if (!mPin1.get())
          return (&mPin1);
        if (!mPin2.get())
          return (&mPin2);
        return (0);
      }

    cPin* getPin(unsigned int i)
      {
        if (i == 0)
          return (&mPin1);
        return (&mPin2);
      }

    cPin& pin1()
      { return (mPin1); }

    cPin& pin2()
      { return (mPin2); }

  private:
    unsigned int miIndex;
    cPin mPin1;
    cPin mPin2;
  };

// 8 bridges //

cPin* firstPin(cTestBridge* pBridge, cTestBridge*& it1, unsigned int& iPos)
  {
    if (!(pBridge)) return (0);
    it1 = pBridge;
    iPos = 0;
    return (it1->getPin(iPos));
  }

cPin* nextPin(cTestBridge*& it, unsigned int& iPos)
  {
    if (!(it)) return (0);
    if (!(it->index()))
      return (0);
    iPos++;
    if (iPos > 1)
      {
        if (!(it->index()))
          return (0);

        iPos = 0;
        it++;
      }
    return (it->getPin(iPos));
  }

void rebirthAllPins(cTestBridge* pBridge)
  {
    cTestBridge* pIterator = 0;
    unsigned int iIterator = 0;
    cPin* pCurrent = firstPin(pBridge, pIterator, iIterator);
    while (pCurrent)
      {
        pCurrent->validate();
        pCurrent = nextPin(pIterator, iIterator);
      };
  }


/**
  This method tests the following:
    * stuck at short (gnd or fixed low)    (uncontrollable pin)
    * stuck at short (+3.3V or fixed high) (uncontrollable pin)
    * short circuit between two pins
    * high input capacitance.

  Pins with stuck at short errors will be invalidated so no damage should be occur.
*/
void runTestWithoutJumper(cTestBridge* pBridge, bool boPassJumper)
  {
    scDbg("Testing all pins without jumper\n");
    if (boPassJumper)
      scDbg("*** INFO: Passing jumper...");

    // (1) set all pins to high and check for high
    //     (if a pin holds on low a stuck a short has been detected)
    {
      scDbg("  + setup all pins to high...\n");
      cTestBridge* pIterator = 0;
      unsigned int iIterator = 0;
      cPin* pCurrent = firstPin(pBridge, pIterator, iIterator);
      while (pCurrent)
        {
          pCurrent->set();
          pCurrent = nextPin(pIterator, iIterator);
        };
    }

    {
      scDbg("  + check all pins high...\n");
      cTestBridge* pIterator = 0;
      unsigned int iIterator = 0;
      cPin* pCurrent = firstPin(pBridge, pIterator, iIterator);
      while (pCurrent)
        {
          if (!(pCurrent->state()))
            {
              scDbg("Stack at short (to low) at pin: ");
              scDbg(pCurrent->portName());
              scDbgPrintDec(pCurrent->portIndex());
              scDbg(" (\""); scDbg(pCurrent->comment()); scDbg("\")\n");
              pCurrent->invalidate();
            }
          pCurrent = nextPin(pIterator, iIterator);
        };
    }


    // (2) Switch each pin to low and measure the time needed for measuring low
    //     The time depends on input capacitance of the pin.
    //     (If the time is longer than expirienced a high input capacitance has been detected.
    //      After a timeout a stack at short (to high) has been detected.)
    {
      cTestBridge* pIterator = 0;
      unsigned int iIterator = 0;
      cPin* pCurrent = firstPin(pBridge, pIterator, iIterator);
      while (pCurrent)
        {
          pCurrent->clr();

          unsigned int iCurrentCap = 0;
          while (pCurrent->state())
            {
              iCurrentCap++;
              if (iCurrentCap > 100000)
                break;
            }

          if (iCurrentCap >= 100000)
            {
              scDbg("Stack at short (to high) at pin: ");
              scDbg(pCurrent->portName());
              scDbgPrintDec(pCurrent->portIndex());
              scDbg(" (\""); scDbg(pCurrent->comment()); scDbg("\")\n");
              pCurrent->invalidate();
            }
          else
            if (iCurrentCap > 1)
              {
                scDbg("*** WARNING! High input capacitance at pin: ");
                scDbg(pCurrent->portName());
                scDbgPrintDec(pCurrent->portIndex());
                scDbg(" (\""); scDbg(pCurrent->comment()); scDbg("\")");
                scDbg(" ("); scDbgPrintDec(iCurrentCap); scDbg(" cycles)\n");
              }
          pCurrent = nextPin(pIterator, iIterator);
        }
    }

    {
      scDbg("  + setup all pins to high...\n");
      unsigned int iBridgeIndex = 0;
      do {
        pBridge[iBridgeIndex].setAll();
        iBridgeIndex++;
      } while (pBridge[iBridgeIndex].index());
    }

    // (3) Check for independence of each pin.
    //     All pins must be set to high (allready set) than each pin will be switched to low.
    //     If another pin also measures low a short circuit between two (ore more) pins has been
    //     detected.
    {
      scDbg("  + checking for independence...\n");
      cTestBridge* pIterator = 0;
      unsigned int iIterator = 0;
      cPin* pCurrent = firstPin(pBridge, pIterator, iIterator);
      while (pCurrent)
        {
          if (!(pCurrent->invalid()))
            {
              pCurrent->get();
              unsigned int iTry = 10000;
              while (!pCurrent->state())
                { // wait for charge parasitic capacity in input //
                  if (!(iTry))
                    break;
                  --iTry;
                }
              if (!(iTry))
                {
                  scDbg("Error: Pin ");
                  scDbg(pCurrent->portName());
                  scDbgPrintDec(pCurrent->portIndex());
                  scDbg(" (\""); scDbg(pCurrent->comment()); scDbg("\")");
                  scDbg(" is not switchable!\n");
                  pCurrent->invalidate();
                  continue;
                }


              cTestBridge* pTestIterator = 0;
              unsigned int iTestIterator = 0;
              cPin* pTestCurrent = firstPin(pBridge, pTestIterator, iTestIterator);
              while (pTestCurrent)
                {
                  if (!(pTestCurrent->invalid()))
                    {
                      if (pTestCurrent != pCurrent)
                        {
                          pTestCurrent->clr();
                          if (!pCurrent->state())
                            {
                              if (pTestCurrent != pCurrent)
                                {
                                  bool boPass = false;
                                  // Jumpermode = pass //
                                  if (boPassJumper)
                                    {
                                      // search for pair //
                                      unsigned int iPIndex = 0;
                                      do {
                                        if (&pBridge[iPIndex].pin1() == pTestCurrent)
                                          if (&pBridge[iPIndex].pin2() == pCurrent)
                                            boPass = true;
                                        if (&pBridge[iPIndex].pin2() == pTestCurrent)
                                          if (&pBridge[iPIndex].pin1() == pCurrent)
                                            boPass = true;
                                        iPIndex++;
                                      } while (pBridge[iPIndex].index());
                                    }

                                  if (!boPass)
                                    {
                                      scDbg("**** Error: Short circuit between ");
                                      scDbg(pCurrent->portName());
                                      scDbgPrintDec(pCurrent->portIndex());
                                      scDbg(" (\""); scDbg(pCurrent->comment()); scDbg("\")");
                                      scDbg(" and ");
                                      scDbg(pTestCurrent->portName());
                                      scDbgPrintDec(pTestCurrent->portIndex());
                                      scDbg(" (\""); scDbg(pTestCurrent->comment()); scDbg("\")");
                                      scDbg(" !\n");
                                    }
                                }
                            }
                          pTestCurrent->set();
                        }
                    }
                  pTestCurrent = nextPin(pTestIterator, iTestIterator);
                }

              pCurrent->set();
          }
          pCurrent = nextPin(pIterator, iIterator);
        }
    }
  }






/**
  This test tries to check all pins setting it to a value and measure the same value
  on another (bridged pin; only there). It requires the jumper has been turned on.
*/
void runTestWithJumper(cTestBridge* pBridge)
  {
    scDbg("Test with jumper...\n");

    // (1) set all pins to low
    {
      scDbg("  + setup all pins to low...\n");
      unsigned int iBridgeIndex = 0;
      do {
        pBridge[iBridgeIndex].clrAll();
        iBridgeIndex++;
      } while (pBridge[iBridgeIndex].index());
    }

    {
      enum CheckType { chkPin1Pin2Low, chkPin2Pin1Low, chkPin1Pin2High, chkPin2Pin1High, chkFinish };
      for (unsigned int iCurrentCheck = chkPin1Pin2Low; iCurrentCheck < chkFinish; ++iCurrentCheck)
        {
          unsigned int iBridgeIndex = 0;
          do {
            cPin* pOutPin = 0;
            cPin* pInPin = 0;
            bool boForcedState = false;

            switch (iCurrentCheck)
              {
              case chkPin1Pin2High:
              case chkPin1Pin2Low:
                pInPin = &pBridge[iBridgeIndex].pin2();
                pOutPin = &pBridge[iBridgeIndex].pin1();
               break;
              case chkPin2Pin1High:
              case chkPin2Pin1Low:
                pInPin = &pBridge[iBridgeIndex].pin1();
                pOutPin = &pBridge[iBridgeIndex].pin2();
               break;

              default: ;
              }

            if ((iCurrentCheck == chkPin1Pin2High) || (iCurrentCheck == chkPin2Pin1High))
              boForcedState = true;

            if ((!(pInPin)) || (!(pOutPin)))
              {
                scDbg("Fatal: No pins defined for check!\n");
                return;
              }

            pInPin->setupInput();
            pOutPin->set(boForcedState);

            bool boPassed1 = false;
            for (unsigned int i = 0; i < 1000; i++)
              {
                boPassed1 = false;
                if (pInPin->get() == boForcedState)
                  {
                    boPassed1 = true;
                    break;
                  }
              }

            if (!(boPassed1))
              {
                scDbg("Pin ");
                scDbg(pBridge[iBridgeIndex].pin1().portName());
                scDbgPrintDec(pBridge[iBridgeIndex].pin1().portIndex());
                scDbg(" (\""); scDbg(pBridge[iBridgeIndex].pin1().comment()); scDbg("\")");
                scDbg(" or ");
                scDbg(pBridge[iBridgeIndex].pin2().portName());
                scDbgPrintDec(pBridge[iBridgeIndex].pin2().portIndex());
                scDbg(" (\""); scDbg(pBridge[iBridgeIndex].pin2().comment()); scDbg("\")");
                scDbg(" seems not to be connected while checking ");
                switch (iCurrentCheck)
                  {
                  case chkPin1Pin2Low:
                     scDbg("pin1 out -> pin2 in (low)");
                   break;
                  case chkPin2Pin1Low:
                     scDbg("pin1 in <- pin2 out (low)");
                   break;
                  case chkPin1Pin2High:
                     scDbg("pin1 out -> pin2 in (high)");
                   break;
                  case chkPin2Pin1High:
                     scDbg("pin1 in <- pin2 out (high)");
                   break;

                  default:
                    scDbg("unknown / bug");
                   break;
                  }
                scDbg("\n");
              }
/*            else
              { scDbg("Passed check ");  scDbgPrintDec(iCurrentCheck); scDbg("\n"); }*/
            pOutPin->set();
            pInPin->set();
            iBridgeIndex++; // next please //
          } while (pBridge[iBridgeIndex].index());
       }
    }

  }

extern "C" void scInitStage0()
  { }

void boot()
  {
    sc_sti();
    sc_waitTime(200);

    // jumper you have to bridge in jumper mode //
    cTestBridge testBridgeJ11[] = { cTestBridge(1, "J 11", GPIO_PD(8),   GPIO_PD(13)),
                                    cTestBridge(1, "J 11", GPIO_PD(10),  GPIO_PD(15)),
                                    cTestBridge(1, "J 11", GPIO_PD(14),  GPIO_PD(7)),
                                    cTestBridge(1, "J 11", GPIO_PD(16),  GPIO_PD(17)),
                                    cTestBridge(1, "J 11", GPIO_PD(18),  GPIO_PD(23)),
                                    cTestBridge(1, "J 11", GPIO_PD(27),  GPIO_PD(25)),
                                    cTestBridge(1, "J 11", GPIO_PD(22),  GPIO_PD(24)),
                                    cTestBridge(1, "J 11", GPIO_PD(21),  GPIO_PD(26)),

                                    cTestBridge(1, "J 10", GPIO_PD(4),   GPIO_PD(1)),
                                    cTestBridge(1, "J 10", GPIO_PD(0),   GPIO_PD(5)),
                                    cTestBridge(1, "J 10", GPIO_PD(2),   GPIO_PD(3)),
                                    cTestBridge(1, "J 10", GPIO_PC(15),   GPIO_PC(5)),
                                    cTestBridge(1, "J 10", GPIO_PC(3),   GPIO_PC(0)),
                                    cTestBridge(1, "J 10", GPIO_PC(2),   GPIO_PC(1)),
                                   // ToDo SPI - lines...

                                    cTestBridge(1, "J 8", GPIO_PB(16),   GPIO_PB(15)),
                                    cTestBridge(1, "J 8", GPIO_PB(24),   GPIO_PB(22)),
                                    cTestBridge(1, "J 8", GPIO_PB(25),   GPIO_PB(19)),
                                    cTestBridge(1, "J 8", GPIO_PB(26),   GPIO_PB(14)),
                                    cTestBridge(1, "J 8", GPIO_PB(27),   GPIO_PB(23)),
                                    cTestBridge(1, "J 8", GPIO_PB(29),   GPIO_PB(18)),
                                    cTestBridge(1, "J 8", GPIO_PB(17),   GPIO_PB(21)),
//                                    cTestBridge(1, "J 8", GPIO_PB(11),   GPIO_PB(13)),
//                                    cTestBridge(1, "J 8", GPIO_PB(12),   GPIO_PB(15)),
                                    cTestBridge(1, "J 8", GPIO_PB(6),   GPIO_PB(9)),
                                    cTestBridge(1, "J 8", GPIO_PB(1),   GPIO_PB(7)),
                                    cTestBridge(1, "J 8", GPIO_PB(0),   GPIO_PB(8)),
                                    cTestBridge(1, "J 8", GPIO_PA(20),   GPIO_PA(22)),
                                    cTestBridge(1, "J 8", GPIO_PA(24),   GPIO_PA(23)),

                                    cTestBridge(0, "xxx", GPIO_PA(0),   GPIO_PA(0)),
                                   };


    showHelp();




    sc_cli();
    while (1)
      {
        if (scDbgHasChar())
          {
            char b = scDbgReadChar(1);
            if (b == 13)
              {
                showHelp();
              }

            if (b == 'W')
              {
                rebirthAllPins(testBridgeJ11);
                runTestWithoutJumper(testBridgeJ11, true);
              }

            if (b == 'J')
              {
                rebirthAllPins(testBridgeJ11);
                runTestWithoutJumper(testBridgeJ11, false);
              }

            if (b == 'w')
              {
                rebirthAllPins(testBridgeJ11);
                runTestWithJumper(testBridgeJ11);
              }
       }
      }
    while (1);
  }
