/*************************************************************************/
  .text
  .align
  .globl __switch_cpu_clock

#define AT91C_BASE_PMC           0xFFFFFC00 /* (PMC) Base Address */
#define PMC_MCKR                 0x30       /* Master Clock Register */
#define PMC_SR                   0x68       /* Status Register */
#define PMC_MOR                  0x20       /* Main Oscillator Register */
#define PMC_SCER                 0x00
#define PMC_SCDR                 0x04
#define PMC_PLLAR                0x28       /* PLL A Register */
#define PMC_PLLBR                0x2C       /* PLL B Register */

#define AT91C_PMC_MOSCS          0x1        /* (PMC) MOSC Status/Enable/Disable/Mask */
#define AT91C_PMC_MCKRDY         0x8        /* (PMC) MCKRDY -//- */
#define AT91C_PMC_LOCKA          0x2        /* (PMC) PLL A Status/Enable/Disable/Mask */
#define AT91C_PMC_LOCKB          0x4        /* (PMC) PLL B Status/Enable/Disable/Mask */


#define AT91C_BASE_SDRAMC        0xFFFFFF90
#define AT91C_SDRAMC_SRR         0x0c       // SelfRefresh register (SDRAM controller)
#define AT91C_SDRAMC_TR          0x04        // Refresh time base register (SDRAM controller)

// debug //
#define AT91C_BASE_PIOD          0xFFFFFA00
#define AT91C_PIOD_CODR          0x34


#define ASYNC_BUS_MODE
//#define SYNC_BUS_MODE
// #define DISABLE_ICACHE


// CP15 control register
/* b0  = 0 - MMU disable
* b1  = 1 - Fault checking disable
* b2  = 0 - DCache disable
* b3:6 - Reserved (1111b)
* b7  = 0 - Little-endian
* b8  = 0 - Sys protection
* b9  = 0 - ROM protection
* b10:11 - Reserved (00b)
* b12 = 1 - ICache enable
* b13 = 0 - Exception registers
*             at 0x00000000
* b14 = 1 - Round Robin replacement
* b15:29 - Reserved (000000000000000b)
* b30:31 = 11 - Asynchronous bus
*                mode
*/


#ifdef ASYNC_BUS_MODE
  #ifdef DISABLE_ICACHE
    #define CFG_P15_C1  0xc0004078
  #else
    #define CFG_P15_C1  0xc0005078
  #endif
#else
  #ifdef SYNC_BUS_MODE
    #define CFG_P15_C1  0x40005078
  #else
    #define CFG_P15_C1  0x00001078
  #endif
#endif



#include "../sc/sc_config.h"

/**** CLOCK SWITCHING ****/
//
// Clock switching must be done with the following steps:
//   * Turn off all interrupts.
//   * Copy the switch_cpu_clock function to SRAM or place it there.
//   * Set SDRAM to self refresh mode.
//   * Switch CPU clock source to slow-clock
//   * Switch PLL frequencies.
//   * Switch CPU clock source to PLLA clock
//   * Setup new value used for SDRAM self refresh
//   * Return to address in SDRAM (this resumes the SDRAM from self refresh)
//   * You may re-enable interrupts now
//
// Note:
//  It's important placing this function in SRAM and disable interrupts because
//  the SDRAM must perform self refresh cycles while reconfiguring the PLLs!
//  An interrupt would stop the SDRAM self refresh (data would lost because without MCK no refresh can generated).
//

//
// Switches the PLLs to the given values.
// Precondition:
//   Interrupts off!
//   The pres field in MCKR must be zero.
//   The processor clock (fosc) must be initialized.
//     (Booth requirements are fullfilled after calling __setup_cpu_clock(PLLA, PLLB);
//
// r0 - PLLA config
// r1 - PLLB config
// r2 - ClockDivSettings
// r3 - new value for SDRAM refresh
//
__switch_cpu_clock:
      stmfd sp!, {r0-r6}

      mov r6, r3 // r6 = SDRAM refresh rate
      mov r5, r2 // clock div settings
      mov r4, r1 // r4 = PLLB - Config
	    mov r3, r0 // r3 = PLLA - Config


// Enter SDRAM self refresh mode...
      ldr r0, =0x1
      ldr r1, =AT91C_BASE_SDRAMC
      str r0, [r1, #AT91C_SDRAMC_SRR]
// From now NO SDRAM access MUST occur until all clocks are switched.


// Setup the following CPU configuration while changing the PLLs
//   Fastbus mode
//   Set MMU disable
//   Fault checking disable
//   DCache disable
//   Little-endian
//   No Sys protection
//   No ROM protection
//   ICache disabled
//   Exception registers at 0x0
//   Random replacement
            ldr      r0, =0x00000078
// write r0 in cp15 control register (cp15:r1)
            mcr      p15, 0, r0, c1, c0,0

// Base address PMC
            ldr      r1, =AT91C_BASE_PMC

// Because the prescaler has been set to zero before
// We have to change the clock source only (CSS fields)
            ldr      r0, =0
            str      r0, [r1, #PMC_MCKR]

Loop2Osc:
// Wait for Master Clock stabilization
            ldr      r0, [r1, #PMC_SR]
            tsts     r0, #AT91C_PMC_MCKRDY
            beq      Loop2Osc

// Turn off processor clock
            ldr      r0, =0x00000001      // bit0 = 1 - processor clock
            str      r0, [r1, #PMC_SCDR]
            ldr      r0, =0x0
            str      r0, [r1, #PMC_MOR]

// Write dirty values to PLLA and PLLB (see errata)
            ldr      r0, =0x2000000
            str      r0, [r1, #PMC_PLLAR]
            ldr      r0, =0
            str      r0, [r1, #PMC_PLLBR]

// Enable processor clock as configured in PLLA and PLLB config.

// Prepare the value to be written to enable main clock
            ldr      r0, =0x0000FF01   // bit0  = 1 - enable main oscill
                                       // byte1 = FF - time slice before
                                       // enabling. In slow ticks/8.
                                       //

            str      r0, [r1, #PMC_MOR]

//
//  Wait while main oscillator have started up
//  i.e. the bit MOSCS in PMC_SR have been set
//  it will be set when number specified in CKGR_MOR as time slice
//  will reach the 0 value
//
Loop3Osc:
// Read state register PMC_SR
            ldr      r0, [r1, #PMC_SR]
            tsts     r0, #AT91C_PMC_MOSCS
            beq      Loop3Osc

// Setup PLLA
            str      r3, [r1, #PMC_PLLAR]

// Wait while PLL will pulled-up.
Loop2PLLA:
// Read state register PMC_SR
            ldr      r0, [r1, #PMC_SR]
            tsts     r0, #AT91C_PMC_LOCKA
            beq      Loop2PLLA

// Setup PLLB.
            str      r4, [r1, #PMC_PLLBR]

// Wait while PLLB will pulled-up.
Loop2PLLB:
// Read state register PMC_SR
            ldr      r0, [r1, #PMC_SR]
            tsts     r0, #AT91C_PMC_LOCKB
            beq      Loop2PLLB

// Setup new value for SDRAM refresh
            ldr      r1, =AT91C_BASE_SDRAMC
            str      r6, [r1, #AT91C_SDRAMC_TR]

// Base address PMC
            ldr      r1, =AT91C_BASE_PMC
//
// Set master clock to real speed.
// We have not changed the PRES field so we can do it with one step.
//
            str      r5, [r1, #PMC_MCKR]

Loop41Osc:
// Wait for master clock stabilization
            ldr      r0, [r1, #PMC_SR]
            tsts     r0, #AT91C_PMC_MCKRDY
            beq      Loop41Osc

// Setup P15:C1
            ldr      r0, =CFG_P15_C1
            mcr      p15, 0, r0, c1, c0,0

// Enable processor clock
            ldr      r0, =0x00000001      // bit0 = 1 - processor clock */
            str      r0, [r1, #PMC_SCER]

// Loop for compensate oscillator start up
            ldr      r0, =0xFFFFFF
Loop5Osc:
            subs     r0, r0, #1
            bhi      Loop5Osc

// wake up sdram
            ldr      r1, =0x20000000
            ldr      r0, [r1]
            ldr      r0, [r1]
            ldr      r0, [r1]
            ldr      r0, [r1]
            ldr      r0, [r1]
            ldr      r0, [r1]
      ldmfd sp!, {r0-r6}
   mov pc, lr
