#include <sc/sc.h>
#include <sc/sc_config.h>
#include <sc/sys_time.h>
#include <at/include/at91rm9200-def.h>
#include "spi_dataflash.h"


#define AT91C_SPI_PCS0_SERIAL_DATAFLASH     0xE     /* Chip Select 0 : NPCS0 %1110 */
#define AT91C_SPI_PCS3_DATAFLASH_CARD       0x7     /* Chip Select 3 : NPCS3 %0111 */
#define AT91C_DATAFLASH_TIMEOUT             20000   /* For AT91F_DataFlashWaitReady */

// << SPI dataflash command set >> //
// READ COMMANDS
#define DB_CONTINUOUS_ARRAY_READ    0xE8    /* Continuous array read */
#define DB_BURST_ARRAY_READ         0xE8    /* Burst array read */
#define DB_PAGE_READ                0xD2    /* Main memory page read */
#define DB_BUF1_READ                0xD4    /* Buffer 1 read */
#define DB_BUF2_READ                0xD6    /* Buffer 2 read */
#define DB_STATUS                   0xD7    /* Status Register */

// PROGRAM and ERASE COMMANDS
#define DB_BUF1_WRITE               0x84    /* Buffer 1 write */
#define DB_BUF2_WRITE               0x87    /* Buffer 2 write */
#define DB_BUF1_PAGE_ERASE_PGM      0x83    /* Buffer 1 to main memory page program with built-In erase */
#define DB_BUF1_PAGE_ERASE_FASTPGM  0x93    /* Buffer 1 to main memory page program with built-In erase, Fast program */
#define DB_BUF2_PAGE_ERASE_PGM      0x86    /* Buffer 2 to main memory page program with built-In erase */
#define DB_BUF2_PAGE_ERASE_FASTPGM  0x96    /* Buffer 1 to main memory page program with built-In erase, Fast program */
#define DB_BUF1_PAGE_PGM            0x88    /* Buffer 1 to main memory page program without built-In erase */
#define DB_BUF1_PAGE_FASTPGM        0x98    /* Buffer 1 to main memory page program without built-In erase, Fast program */
#define DB_BUF2_PAGE_PGM            0x89    /* Buffer 2 to main memory page program without built-In erase */
#define DB_BUF2_PAGE_FASTPGM        0x99    /* Buffer 1 to main memory page program without built-In erase, Fast program */
#define DB_PAGE_ERASE               0x81    /* Page Erase */
#define DB_BLOCK_ERASE              0x50    /* Block Erase */
#define DB_PAGE_PGM_BUF1            0x82    /* Main memory page through buffer 1 */
#define DB_PAGE_FASTPGM_BUF1        0x92    /* Main memory page through buffer 1, Fast program */
#define DB_PAGE_PGM_BUF2            0x85    /* Main memory page through buffer 2 */
#define DB_PAGE_FastPGM_BUF2        0x95    /* Main memory page through buffer 2, Fast program */

// ADDITIONAL COMMANDS
#define DB_PAGE_2_BUF1_TRF          0x53    /* Main memory page to buffer 1 transfert */
#define DB_PAGE_2_BUF2_TRF          0x55    /* Main memory page to buffer 2 transfert */
#define DB_PAGE_2_BUF1_CMP          0x60    /* Main memory page to buffer 1 compare */
#define DB_PAGE_2_BUF2_CMP          0x61    /* Main memory page to buffer 2 compare */
#define DB_AUTO_PAGE_PGM_BUF1       0x58    /* Auto page rewrite throught buffer 1 */
#define DB_AUTO_PAGE_PGM_BUF2       0x59    /* Auto page rewrite throught buffer 2 */



/**
  \brief Initializes the SPI hardware.
*/
void spiDFInit(unsigned int iMCK, unsigned int iSpeed)
  {
    // turn off spi...
    AT91C_BASE_SPI->SPI_MR = 0;
    AT91C_BASE_SPI->SPI_CR = (1 << 1);
    // Configure PIOs
    AT91C_BASE_PIOA->PIO_ASR = AT91C_PA4_NPCS1 | AT91C_PA1_MOSI | AT91C_PA5_NPCS2 |
                   AT91C_PA6_NPCS3 | AT91C_PA0_MISO | AT91C_PA2_SPCK | AT91C_PA3_NPCS0;
    AT91C_BASE_PIOA->PIO_PDR = AT91C_PA4_NPCS1 | AT91C_PA1_MOSI | AT91C_PA5_NPCS2 |
                   AT91C_PA6_NPCS3 | AT91C_PA0_MISO | AT91C_PA2_SPCK | AT91C_PA3_NPCS0;

    // Enable clock
    AT91C_BASE_PMC->PMC_PCER = 1 << AT91C_ID_SPI;

    // turn on spi...
    AT91C_BASE_SPI->SPI_CR = (1 << 0);
    // Reset the SPI...
    AT91C_BASE_SPI->SPI_CR = AT91C_SPI_SWRST;

    /* Configure SPI in Master Mode with No CS selected !!! */
    AT91C_BASE_SPI->SPI_MR = AT91C_SPI_MSTR | AT91C_SPI_MODFDIS | AT91C_SPI_PCS;

    // Configure SPI in Master Mode with No CS selected...
    // Configure CS0 (dataflash)
//    *(AT91C_SPI_CSR + 0) = AT91C_SPI_CPOL | ((AT91C_MASTER_CLOCK / (2*10000000)) << 8) | (44 << 16) | (2 << 24);
    *(AT91C_SPI_CSR + 0) = AT91C_SPI_CPOL | ((iMCK / (2*iSpeed)) << 8) | (44 << 16) | (2 << 24);
//    *(AT91C_SPI_CSR + 3) = AT91C_SPI_CPOL | (AT91C_SPI_DLYBS & 0x100000) | ((AT91C_MASTER_CLOCK / (2*iSpeed)) << 8) | (150 << 16);
  }



/**
  \brief Enable SPI access for dataflash.
*/
void spiDFEnable(unsigned int iCS)
  {
    AT91C_BASE_SPI->SPI_MR &= 0xfff0ffff;
    AT91C_BASE_SPI->SPI_MR |= ((AT91C_SPI_PCS0_SERIAL_DATAFLASH<<16) & AT91C_SPI_PCS);
    AT91C_BASE_SPI->SPI_CR = AT91C_SPI_SPIEN;
  }

/**
  \brief Write to spi bus.
  \return Returns 0 on error and (1==1 on success).
*/
unsigned int spiWrite(struct tDFCmdDesc* pDesc)
  {
    AT91C_BASE_SPI->SPI_PTCR = AT91C_PDC_TXTDIS + AT91C_PDC_RXTDIS;
    /* Initialize the Transmit and Receive Pointer */
    AT91C_BASE_SPI->SPI_RPR = (unsigned int)pDesc->rx_cmd_pt ;
    AT91C_BASE_SPI->SPI_TPR = (unsigned int)pDesc->tx_cmd_pt ;

    /* Intialize the Transmit and Receive Counters */
    AT91C_BASE_SPI->SPI_RCR = pDesc->rx_cmd_size;
    AT91C_BASE_SPI->SPI_TCR = pDesc->tx_cmd_size;

    if ( pDesc->tx_data_size != 0 )
      {
        /* Initialize the Next Transmit and Next Receive Pointer */
        AT91C_BASE_SPI->SPI_RNPR = (unsigned int)pDesc->rx_data_pt ;
        AT91C_BASE_SPI->SPI_TNPR = (unsigned int)pDesc->tx_data_pt ;

        /* Intialize the Next Transmit and Next Receive Counters */
        AT91C_BASE_SPI->SPI_RNCR = pDesc->rx_data_size ;
        AT91C_BASE_SPI->SPI_TNCR = pDesc->tx_data_size ;
      }



    AT91C_BASE_SPI->SPI_PTCR = AT91C_PDC_TXTEN + AT91C_PDC_RXTEN;

    while (!(AT91C_BASE_SPI->SPI_SR & AT91C_SPI_RXBUFF))
      {
//        scDbg("0x"); scDbgPrintHex32(AT91C_BASE_SPI->SPI_RCR); scDbg("\n");
      }

    AT91C_BASE_SPI->SPI_PTCR = AT91C_PDC_TXTDIS + AT91C_PDC_RXTDIS;
    return (1==1);
  }


/**
  \brief Submit a command to dataflash.
  \return Returns 0 on error and (1==1) on success.
*/
unsigned int spiDFSendCommand(struct tDFDevice* pDataFlash, unsigned char OpCode, unsigned int CmdSize, unsigned int DataflashAddress)
  {
    // Process the address to obtain page address and byte address.
    unsigned int adr = ((DataflashAddress / (pDataFlash->miPageSize)) << pDataFlash->miPageOffset) + (DataflashAddress % (pDataFlash->miPageSize));

    // Setup command and buffer...
    pDataFlash->mCmd.command[0] = OpCode;
    if (pDataFlash->miNumPages >= 16384)
      {
        pDataFlash->mCmd.command[1] = (unsigned char)((adr & 0x0F000000) >> 24);
        pDataFlash->mCmd.command[2] = (unsigned char)((adr & 0x00FF0000) >> 16);
        pDataFlash->mCmd.command[3] = (unsigned char)((adr & 0x0000FF00) >> 8);
        pDataFlash->mCmd.command[4] = (unsigned char)(adr & 0x000000FF);
      }
    else
      {
        pDataFlash->mCmd.command[1] = (unsigned char)((adr & 0x00FF0000) >> 16);
        pDataFlash->mCmd.command[2] = (unsigned char)((adr & 0x0000FF00) >> 8);
        pDataFlash->mCmd.command[3] = (unsigned char)(adr & 0x000000FF) ;
        pDataFlash->mCmd.command[4] = 0;
      }
    pDataFlash->mCmd.command[5] = 0;
    pDataFlash->mCmd.command[6] = 0;
    pDataFlash->mCmd.command[7] = 0;

    /* Initialize the SpiData structure for the spi write fuction */
    pDataFlash->mCmd.tx_cmd_pt   = pDataFlash->mCmd.command;
    pDataFlash->mCmd.tx_cmd_size = CmdSize;
    pDataFlash->mCmd.rx_cmd_pt   = pDataFlash->mCmd.command;
    pDataFlash->mCmd.rx_cmd_size = CmdSize;

    return (spiWrite(&pDataFlash->mCmd));
  }


/**
  \brief Read dataflash status.
  \return Returns 0 on error, (1==1) on success.
  The variable pDesc->miDFState will be updated accordly.
*/
unsigned int spiDFGetState(struct tDFCmdDesc* pDesc)
  {
    unsigned int iResult = 0;
    pDesc->command[0] = DB_STATUS;
    pDesc->command[1] = 0;

    pDesc->tx_data_size = 0;
    pDesc->tx_cmd_pt = pDesc->command;
    pDesc->rx_cmd_pt = pDesc->command;
    pDesc->rx_cmd_size = 2;
    pDesc->tx_cmd_size = 2;
    iResult = spiWrite(pDesc);
    pDesc->DataFlash_state = *((unsigned char *) (pDesc->rx_cmd_pt)+1);
    return (iResult);
  }

/**
  \brief Wait until the dataflash is ready or the timeout has been exceeded.
*/
unsigned int spiDFWaitReady(struct tDFCmdDesc* pDesc, unsigned int iTimeOut)
  {
    pDesc->DataFlash_state = 0;
    do
      {
        // TODO timeout !
        spiDFGetState(pDesc);
        if ((pDesc->DataFlash_state & 0x80) == 0x80)
          return (1);
      } while(1);

    return (0);
  }


/**
  \brief Setup continous read.
  \param pDevice - the flash device
  \param iSrc - the flash source address
  \param pDst - the buffer address
  \param iSize - the amount of bytes to read
  \return Returns zero on errors and (1==1 for success.
*/
unsigned int spiDFSetupContinuousRead(struct tDFDevice* pDevice, unsigned int iSrc, char* pDst, unsigned int iSize)
  {
    if ((iSrc+iSize) > pDevice->miPageSize*pDevice->miNumPages)
      return (0); // out of range //

    pDevice->mCmd.rx_data_pt = pDst;
    pDevice->mCmd.rx_data_size = iSize;
    pDevice->mCmd.tx_data_pt = pDst;
    pDevice->mCmd.tx_data_size = iSize;

    return (spiDFSendCommand(pDevice, DB_CONTINUOUS_ARRAY_READ, 8, iSrc));
  }


/**
  \brief Copies data from S(D)RAM to SRAM page in dataflash.
  \param pDataFlash - the flash device
  \param bCmd - command to execute
  \param iPage - page (0/1)
*/
// AT91F_MainMemoryToBufferTransfert
unsigned int spiDFCopyToDFSRAM(struct tDFDevice* pDataFlash, unsigned char bCmd, unsigned int iPage)
  {
    unsigned int iCmdSize = (pDataFlash->miNumPages >= 16384) ? (5):(4);

    // Test if the buffer command is legal
    if ((bCmd != DB_PAGE_2_BUF1_TRF) && (bCmd != DB_PAGE_2_BUF2_TRF))
      return 0; // invalid request //

    // no data to transmit or receive
    pDataFlash->mCmd.tx_data_size = 0;

    return(spiDFSendCommand(pDataFlash, bCmd, iCmdSize, iPage*pDataFlash->miPageSize));
  }



/**
  \brief Writes data to the flash's internal SRAM buffer.
  \param pDevice - the flash device
  \param bCmd - the command to use (page selector)
  \param pSrc - the source of data in S(D)RAM
  \param iOffset - offset inside SRAM page
  \param iSize - the amount of bytes to write
*/
// AT91F_DataFlashWriteBuffer
unsigned int spiDFWriteSDRAMPage(struct tDFDevice* pDataFlash, unsigned char bCmd, char* pSrc, unsigned int iBfAddr, unsigned int iSize)
  {
    unsigned int iCmdSize = 0;

    /* Test if the buffer command is legal */
    if ((bCmd != DB_BUF1_WRITE) && (bCmd != DB_BUF2_WRITE))
      return 0;

    /* buffer address must be lower than page size */
    if (iBfAddr > pDataFlash->miPageSize)
      return 0;


    pDataFlash->mCmd.command[0] = bCmd;
    pDataFlash->mCmd.command[1] = 0;

    iCmdSize = 4;
    pDataFlash->mCmd.command[2] = (unsigned char)(((unsigned int)(iBfAddr &  pDataFlash->miByteMask)) >> 8);
    pDataFlash->mCmd.command[3] = (unsigned char)((unsigned int)iBfAddr  & 0x00FF);
    pDataFlash->mCmd.command[4] = 0;

    if (pDataFlash->miNumPages >= 16384)
      {
        pDataFlash->mCmd.command[2] = 0;
        pDataFlash->mCmd.command[3] = (unsigned char)(((unsigned int)(iBfAddr & pDataFlash->miByteMask)) >> 8);
        pDataFlash->mCmd.command[4] = (unsigned char)((unsigned int)iBfAddr & 0x00FF);
        iCmdSize = 5;
      }

    pDataFlash->mCmd.tx_cmd_pt    = pDataFlash->mCmd.command;
    pDataFlash->mCmd.tx_cmd_size  = iCmdSize;
    pDataFlash->mCmd.rx_cmd_pt    = pDataFlash->mCmd.command;
    pDataFlash->mCmd.rx_cmd_size  = iCmdSize;

    pDataFlash->mCmd.rx_data_pt   = pSrc;
    pDataFlash->mCmd.tx_data_pt   = pSrc;
    pDataFlash->mCmd.rx_data_size = iSize;
    pDataFlash->mCmd.tx_data_size = iSize;

    return spiWrite(&pDataFlash->mCmd);
  }



/**
  \brief Erase an entire page.
  \param pDevice - the flash
  \param iPage - the page
*/
// AT91F_PageErase
unsigned int spiDFErasePage(struct tDFDevice* pDataFlash, unsigned int iPage)
  {
    unsigned int iCmdSize = (pDataFlash->miNumPages >= 16384) ? (5):(4);
    pDataFlash->mCmd.tx_data_size = 0;
    return (spiDFSendCommand(pDataFlash, DB_PAGE_ERASE, iCmdSize, iPage*pDataFlash->miPageSize));
  }



/**
  \brief Readout the SRAM buffer and write it to the flash-array.
  \param bCmd - page selector (SRAM page 0/1)
*/
// AT91F_WriteBufferToMain
unsigned int spiDfWriteSRAMToFlash(struct tDFDevice* pDataFlash, unsigned char bCmd, unsigned int pDest)
  {
    unsigned int iCmdSize = (pDataFlash->miNumPages >= 16384) ? (5):(4);
    /* Test if the buffer command is correct */
    if ((bCmd != DB_BUF1_PAGE_PGM) &&
        (bCmd != DB_BUF1_PAGE_ERASE_PGM) &&
        (bCmd != DB_BUF2_PAGE_PGM) &&
        (bCmd != DB_BUF2_PAGE_ERASE_PGM))
          return 0;

    /* no data to transmit or receive */
    pDataFlash->mCmd.tx_data_size = 0;

    /* Send the command to the dataflash */
    return (spiDFSendCommand(pDataFlash, bCmd, iCmdSize, pDest));
}


/**
  \brief Patch a page.
  \param pSrc - source data buffer
  \param iDst - byte in dataflash
  \param iSize - number of bytes to transfer
*/
// AT91F_PartialPageWrite
unsigned int spiDFPartialPageWrite(struct tDFDevice* pDataFlash, char* pSrc, unsigned int iDst, unsigned int iSize)
  {
    unsigned int iPage = iDst / (pDataFlash->miPageSize);
    unsigned int iAdrInPage = iDst % (pDataFlash->miPageSize);

    /* Read the contents of the page in the Sram Buffer */
    if (!(spiDFCopyToDFSRAM(pDataFlash, DB_PAGE_2_BUF1_TRF, iPage)))
      return (0);
    if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
      return (0);

    /*Update the SRAM buffer */
    if (!(spiDFWriteSDRAMPage(pDataFlash, DB_BUF1_WRITE, pSrc, iAdrInPage, iSize)))
      return (0);
    if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
      return (0);

    /* Erase page if a 128 Mbits device */
    if (pDataFlash->miNumPages >= 16384)
      {
        if (!(spiDFErasePage(pDataFlash, iPage)))
          return (0);
        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);
      }

    // Rewrite the modified Sram Buffer in the main memory
    return (spiDfWriteSRAMToFlash(pDataFlash, DB_BUF1_PAGE_ERASE_PGM, (iPage*pDataFlash->miPageSize)));
  }



/**
  \brief Writes the given amount of bytes to the flash.
*/
unsigned int spiDFWrite(struct tDFDevice* pDataFlash, char* pSrc, unsigned int iDst, unsigned int iSize)
  {
    spiDFEnable(pDataFlash->miCS);

    if ((iDst + iSize) > (pDataFlash->miPageSize*(pDataFlash->miNumPages)))
      return 0;

    // If destination does not fit a page start address
    if ((iDst % ((pDataFlash->miPageSize))) != 0)
      {
        unsigned int iLength = pDataFlash->miPageSize - (iDst % ((pDataFlash->miPageSize)));

        if (iSize < iLength)
            iLength = iSize;

        if (!(spiDFPartialPageWrite(pDataFlash, pSrc, iDst, iLength)))
          return (0);

        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);

        iSize -= iLength;
        iDst += iLength;
        pSrc += iLength;
      }

    while ((iSize >= pDataFlash->miPageSize))
      {
        // Program dataflash page.
        unsigned int iPage = iDst / (pDataFlash->miPageSize);
        if (!(spiDFWriteSDRAMPage(pDataFlash, DB_BUF1_WRITE, pSrc, 0, pDataFlash->miPageSize)))
          return (0);
        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);

        if (!(spiDFErasePage(pDataFlash, iPage)))
          return (0);
        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);

        if (!(spiDfWriteSRAMToFlash(pDataFlash, DB_BUF1_PAGE_PGM, iDst)))
          return (0);

        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);

        /* Update size, source and destination pointers */
        iSize -= pDataFlash->miPageSize;
        iDst += pDataFlash->miPageSize;
        pSrc += pDataFlash->miPageSize;
      }

    // If still some bytes to read
    if (iSize > 0)
      {
        if (!(spiDFPartialPageWrite(pDataFlash, pSrc, iDst, iSize)))
          return (0);
        if (!(spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT)))
          return (0);
      }
    return (1==1);
  }



/**
  \brief Reads the given amount of bytes from flas.
*/
unsigned int spiDFRead(struct tDFDevice* pDataFlash, unsigned int iSrc, char* pDst, unsigned int iSize)
  {
    unsigned int iBlockSize = 0x8000;

    spiDFEnable(pDataFlash->miCS);

    spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT);

    while (iSize)
      {
        unsigned int iSizeToRead = (iSize < iBlockSize) ? (iSize):(iBlockSize);

        if (!(spiDFSetupContinuousRead(pDataFlash, iSrc, pDst, iSizeToRead)))
          return (0);
        spiDFWaitReady(&pDataFlash->mCmd, AT91C_DATAFLASH_TIMEOUT);

        iSize -= iSizeToRead;
        iSrc += iSizeToRead;
        pDst += iSizeToRead;
      }
    return (1==1);
  }


/**
  \brief Probes dataflash.
  \return Returns 0 on errors and values != 0 on success.
*/
unsigned int spiDFProbe(unsigned int iCS, struct tDFCmdDesc* pDesc)
  {
    spiDFEnable(iCS);
    if (!(spiDFGetState(pDesc)))
      return (0);
    return ((pDesc->command[1] == 0xFF)? 0: (pDesc->command[1] & 0x3C));
  }




#define DF_TYPE_AT45DB161 0x2c
#define DF_TYPE_AT45DB321 0x34
#define DF_TYPE_AT45DB642 0x3c
#define DF_TYPE_AT45DB128 0x10

/**
  \brief Reads information about an attached device.
  \note This method requires that spiDFInit() has been called.
*/
unsigned int spiDFProbeDevice(unsigned int iCS, struct tDFDevice* pInfo)
  {
    pInfo->miID = 0;
    pInfo->miNumPages = 0;
    pInfo->miPageOffset = 0;
    pInfo->miByteMask = 0;
    pInfo->miCS = 0;
    pInfo->miID = 0;

    pInfo->miID = spiDFProbe(iCS, &pInfo->mCmd);
    switch (pInfo->miID)
      {
        case DF_TYPE_AT45DB161:
          pInfo->miNumPages = 4096;
          pInfo->miPageSize = 528;
          pInfo->miPageOffset = 10;
          pInfo->miByteMask = 0x300;
          pInfo->miCS = iCS;
         break;

        case DF_TYPE_AT45DB321:
          pInfo->miNumPages = 8192;
          pInfo->miPageSize = 528;
          pInfo->miPageOffset = 10;
          pInfo->miByteMask = 0x300;
          pInfo->miCS = iCS;
         break;

        case DF_TYPE_AT45DB642:
          pInfo->miNumPages = 8192;
          pInfo->miPageSize = 1056;
          pInfo->miPageOffset = 11;
          pInfo->miByteMask = 0x700;
          pInfo->miCS = iCS;
         break;

        case DF_TYPE_AT45DB128:
          pInfo->miNumPages = 16384;
          pInfo->miPageSize = 1056;
          pInfo->miPageOffset = 11;
          pInfo->miByteMask = 0x700;
          pInfo->miCS = iCS;
         break;
     }
    return (pInfo->miID != 0);
  }

void spiDFDebug()
  {
    scDbg("\nDF-DBG\n");
    scDbg("SPI-Status: 0x"); scDbgPrintHex32(AT91C_BASE_SPI->SPI_SR); scDbg("\n");
    scDbg("SPI-MR: 0x"); scDbgPrintHex32(AT91C_BASE_SPI->SPI_MR); scDbg("\n");
  }
